-- ENCORE MODE MAIN SCRIPT
-- MADE BY DELF / ffoxD.
-- REUSABLE

-- the multiplayer gametype
G_AddGametype({
    name = "Encore",
    identifier = "encore",
    typeoflevel = TOL_COOP,
    rules = GTR_CAMPAIGN|GTR_FRIENDLY|GTR_LIVES|GTR_EMERALDHUNT|GTR_SPAWNENEMIES|GTR_ALLOWEXIT|GTR_CUTSCENES,
    intermissiontype = int_coop,
    headerleftcolor = 35,
    headerrightcolor = 183,
	rankingtype = GT_COOP,
    description = "Start off with one character. Find Certain Monitors to get more or swap between ones you have. Press Custom 3 to switch characters on the fly!"
})

local bots = false -- multiplayer bots toggle

-- recreates the list for the extra character monitor
local function E_RebuildExtraList(player)
	player.extralist = {}
	local sknss = {}

	for i in ipairs(player.encorelives) do -- creates a list of used skins for performance reasons
		sknss[player.encorelives[i].skin] = true
	end

	-- iterates throught all the skins, ignores skins that are invalid or already in the list, and otherwise adds them to the box list.
	for skin in skins.iterate do
		if not skin or not skin.valid then
			continue
		end

		if sknss[skin.name] then
			continue
		end

		if not R_SkinUsable(player, skin.name) then
			continue
		end

		table.insert(player.extralist, skin)
	end
end

-- updates the player and stuff
local function E_Update(player)
	R_SetPlayerSkin(player, player.encorelives[1].skin)
	player.mo.color = player.encorelives[1].color
	E_RebuildExtraList(player) -- the list used by the Character Monitor
	if bots or not multiplayer then
		-- spawn bot
		if #player.encorelives >= 2 then
			if not player.encorebot or not player.encorebot.valid then
				player.encorebot = G_AddPlayer(player.encorelives[2].skin, player.encorelives[2].color, Tails, BOT_2PAI)
			end
			if player.encorebot and player.encorebot.mo and player.encorebot.mo.valid then
				R_SetPlayerSkin(player.encorebot, player.encorelives[2].skin)
				player.encorebot.mo.color = player.encorelives[2].color
			end
		else
			if player.encorebot then
				if player.encorebot.valid then
					G_RemovePlayer(#player.encorebot)
				end
				player.encorebot = nil
			end
		end
	end
end

-- adds a buddy
local function E_AddBuddy(buddylist, buddy)
	buddylist[#buddylist+1] = {
		skin = buddy,
		color = skins[buddy].prefcolor
	}
end

-- kills a buddy
local function E_RemoveBuddy(buddylist, buddynum)
	table.remove(buddylist, buddynum)
end

-- used by the extra character box
function A_ExtraChar(actor)
	local player

	if not actor.target or not actor.target.player then
		print("Powerup has no target.")
		return
	end

	player = actor.target.player

	if actor.type == MT_ENC_ICON and actor.tracer then
		-- We're using the overlay, so use the overlay 1up sprite (no text)
		actor.sprite = SPR_TV1P
	end

	if ultimatemode then --I don't THINK so!
		S_StartSound(player.mo, sfx_lose)
		return
	end

	if (actor.info.seesound)
		S_StartSound(player.mo, actor.info.seesound)
	end
	E_AddBuddy(player.encorelives, actor.sknn)
	E_Update(player)
end

-- Switches between you and the bot (unrelated: it can theoretically switch between 2 players of any kind (not just bots) lol)
local function E_SwitchPlayer(p1, p2)
	local a = {x1 = p1.mo.x, x2 = p2.mo.x, y1 = p1.mo.y, y2 = p2.mo.y, z1 = p1.mo.z, z2 = p2.mo.z, s1 = p1.mo.skin, s2 = p2.mo.skin, pf1 = p1.pflags, pf2 = p2.pflags, c1 = p1.mo.color, c2 = p2.mo.color, sc1 = p1.skincolor, sc2 = p2.skincolor, f1 = p1.mo.frame, f2 = p2.mo.frame}
	P_TeleportMove(p1.mo, a.x2, a.y2, a.z2)
	P_TeleportMove(p2.mo, a.x1, a.y1, a.z1)
	R_SetPlayerSkin(p1, a.s2)
	R_SetPlayerSkin(p2, a.s1)
	p1.mo.color, p2.mo.color = a.c2, a.c1
	p1.skincolor, p2.skincolor = a.sc2, a.sc1
	p1.mo.momx, p2.mo.momx = $2, $1
	p1.mo.momy, p2.mo.momy = $2, $1
	p1.mo.momz, p2.mo.momz = $2, $1
	p1.rmomx, p2.rmomx = $2, $1
	p1.rmomy, p2.rmomy = $2, $1
	p1.pflags, p2.pflags = $2, $1
	if a.pf1 & PF_FLIPCAM then
		p1.pflags = $|PF_FLIPCAM
	else
		p1.pflags = $ & ~PF_FLIPCAM
	end
	if a.pf1 & PF_ANALOGMODE then
		p1.pflags = $|PF_ANALOGMODE
	else
		p1.pflags = $ & ~PF_ANALOGMODE
	end
	if a.pf1 & PF_DIRECTIONCHAR then
		p1.pflags = $|PF_DIRECTIONCHAR
	else
		p1.pflags = $ & ~PF_DIRECTIONCHAR
	end
	if a.pf1 & PF_AUTOBRAKE then
		p1.pflags = $|PF_AUTOBRAKE
	else
		p1.pflags = $ & ~PF_AUTOBRAKE
	end
	if a.pf2 & PF_FLIPCAM then
		p2.pflags = $|PF_FLIPCAM
	else
		p2.pflags = $ & ~PF_FLIPCAM
	end
	if a.pf2 & PF_ANALOGMODE then
		p2.pflags = $|PF_ANALOGMODE
	else
		p2.pflags = $ & ~PF_ANALOGMODE
	end
	if a.pf2 & PF_DIRECTIONCHAR then
		p2.pflags = $|PF_DIRECTIONCHAR
	else
		p2.pflags = $ & ~PF_DIRECTIONCHAR
	end
	if a.pf2 & PF_AUTOBRAKE then
		p2.pflags = $|PF_AUTOBRAKE
	else
		p2.pflags = $ & ~PF_AUTOBRAKE
	end
	p1.panim, p2.panim = $2, $1
	--p1.powers[pw_super], p2.powers[pw_super] = $2, $1
	p1.powers[pw_carry], p2.powers[pw_carry] = $2, $1
	p1.powers[pw_tailsfly], p2.powers[pw_tailsfly] = $2, $1
	p1.powers[pw_underwater], p2.powers[pw_underwater] = $2, $1
	p1.powers[pw_spacetime], p2.powers[pw_spacetime] = $2, $1
	p1.powers[pw_dye], p2.powers[pw_dye] = $2, $1
	p1.powers[pw_flashing], p2.powers[pw_flashing] = $2, $1
	p1.mo.flags, p2.mo.flags = $2, $1
	p1.mo.flags2, p2.mo.flags2 = $2, $1
	p1.mo.eflags, p2.mo.eflags = $2, $1
	p1.dashmode, p2.dashmode = $2, $1
	p1.glidetime, p2.glidetime = $2, $1
	p1.climbing, p2.climbing = $2, $1
	p1.lastlinehit, p2.lastlinehit = $2, $1
	p1.lastsidehit, p2.lastsidehit = $2, $1
	p1.deadtimer, p2.deadtimer = $2, $1
	p1.homing, p2.homing = $2, $1
	p1.skidtimer, p2.skidtimer = $2, $1
	p1.mo.health, p2.mo.health = $2, $1
	p1.mo.friction, p2.mo.friction = $2, $1
	p1.mo.scale, p2.mo.scale = $2, $1
	p1.mo.state, p2.mo.state = $2, $1
	p1.playerstate, p2.playerstate = $2, $1
	--p1.mo.sprite2, p2.mo.sprite2 = $2, $1
	p1.mo.renderflags, p2.mo.renderflags = $2, $1
	p1.mo.frame, p2.mo.frame = a.f2, a.f1
	p1.mo.blendmode, p2.mo.blendmode = $2, $1
	p1.mo.spritexscale, p2.mo.spritexscale = $2, $1
	p1.mo.spriteyscale, p2.mo.spriteyscale = $2, $1
	p1.mo.spritexoffset, p2.mo.spritexoffset = $2, $1
	p1.mo.spriteyoffset, p2.mo.spriteyoffset = $2, $1
	p1.mo.angle, p2.mo.angle = $2, $1
	p1.drawangle, p2.drawangle = $2, $1
end

-- scrambled eggs.
local function E_Shuffle(t)
  local tbl = {}
  for i = 1, #t do
    tbl[i] = t[i]
  end
  for i = #tbl, 2, -1 do
    local j = P_RandomKey(i)+1
    tbl[i], tbl[j] = tbl[j], tbl[i]
  end
  return tbl
end

-- used by the random box
function A_ScrambleBox(actor)
	local player

	if not actor.target or not actor.target.player then
		print("Powerup has no target.")
		return
	end

	player = actor.target.player

	if #player.encorelives > 2 then
		player.encorelives = E_Shuffle(player.encorelives)
		E_Update(player)
		if (actor.info.seesound)
			S_StartSound(player.mo, actor.info.seesound)
		end
		P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_EXPLODE)
	end
end

-- swap 2 buddies one for another.
local function E_Switch(buddylist, indx1, indx2)
	buddylist[indx1], buddylist[indx2] = buddylist[indx2], buddylist[indx1]
end

-- shift the list
local function E_Swap(buddylist)
	table.insert(buddylist, #buddylist, table.remove(buddylist, 1))
end

-- used by the swap monitor
function A_SwapBox(actor)
	local player

	if not actor.target or not actor.target.player then
		print("Powerup has no target.")
		return
	end

	player = actor.target.player

	if #player.encorelives > 2 then
		E_Swap(player.encorelives)
		E_Update(player)
		if (actor.info.seesound)
			S_StartSound(player.mo, actor.info.seesound)
		end
		P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_EXPLODE)
	end
end

-- main function
local function E_Main(player)
	-- detects if you are currently in Encore Mode
	if gametype ~= GT_ENCORE and multiplayer then
		-- WIP (TODO: remove bots, reset everything)
		return
	end
	
	-- basically you don't want all this in NiGHTS (is that how you spell it...?)
	if (maptol & TOL_NIGHTS) then
		return
	end

	-- you are in Encore Mode right now, yay
	-- check if you are ok
	if not (player and player.valid and player.mo and player.mo.valid) then
		return
	end
	if player.bot == BOT_2PAI then -- bots should not, well,
		return
	end
	-- player.encorelives = table/list of all the characters you have.
	if not player.encorelives then
		player.encorelives = {} -- creates it
		 -- each entry of the list is a table too. contains information like the character
		player.encorelives[1] = {
			skin = player.mo.skin,
			color = player.mo.color
		}
		if not multiplayer then
			for p in players.iterate do
				if p.bot == BOT_2PAI then
					player.encorelives[2] = {
						skin = p.mo.skin,
						color = p.mo.color
					}
					player.encorebot = p
				end
			end
		end
		E_Update(player)
	end

	-- ok so like we have our list by now so everything is done right?
	-- well yes but actiually no
	-- switching
	if player.cmd.buttons & BT_CUSTOM3 then
		player.sbutton = $+1 or 1
	else
		player.sbutton = 0
	end
	if player.switchcooldown then
		player.switchcooldown = $-1
		local lockon = P_SpawnMobj(player.mo.x + player.rmomx, player.mo.y + player.rmomy, player.mo.z + player.mo.height + player.mo.momz + player.mo.pmomz, MT_ENSWAP)
		lockon.scale = player.mo.scale
		if player.switchcooldown > TICRATE/3 and player.switchcooldown < 2*TICRATE/3 and player.switchcooldownlast then
			lockon.frame = $|B
		elseif not (player.switchcooldown > TICRATE/3 and player.switchcooldown < 2*TICRATE/3) and not player.switchcooldownlast then
			lockon.frame = $|B
		else
			lockon.frame = $|A
		end
	end

	if player.sbutton == 1 then
		if P_IsObjectOnGround(player.mo) and not player.switchcooldown and player.encorelives[2] then
			if player.encorebot and player.encorebot.valid and P_IsObjectOnGround(player.encorebot.mo) and P_CheckSight(player.mo, player.encorebot.mo) then
				E_SwitchPlayer(player, player.encorebot)
				E_Switch(player.encorelives, 1, 2)
				player.switchcooldown = TICRATE
				if player.switchcooldownlast then
					player.switchcooldownlast = false
				else
					player.switchcooldownlast = true
				end
				E_Update(player)
				S_StartSound(player.mo, sfx_swap, player)
			elseif player.encorebot and player.encorebot.valid then
				S_StartSound(player.mo, sfx_swapf, player)
			else
				E_Switch(player.encorelives, 1, 2)
				player.switchcooldown = TICRATE
				if player.switchcooldownlast then
					player.switchcooldownlast = false
				else
					player.switchcooldownlast = true
				end
				E_Update(player)
				S_StartSound(player.mo, sfx_swap, player)
			end
		else
			S_StartSound(player.mo, sfx_swapf, player)
		end
	end

	-- death handler. it may be a bit buggy...?
	if player.encorebot and player.encorebot.valid and player.deadtimer == 2*TICRATE and not player.encorebot.deadtimer and not player.deadbypit and #player.encorelives > 1 then
		E_Switch(player.encorelives, 1, 2)
		E_SwitchPlayer(player.encorebot, player)
		E_RemoveBuddy(player.encorelives, 2)
		G_RemovePlayer(#player.encorebot)
		player.encorebot = nil
		E_Update(player)
		player.powers[pw_flashing] = TICRATE
		player.lives = #player.encorelives
		player.switchcooldown = TICRATE
		if player.switchcooldownlast then
			player.switchcooldownlast = false
		else
			player.switchcooldownlast = true
		end
	elseif player.deadtimer == 2*TICRATE then
		if #player.encorelives > 1 then
			E_RemoveBuddy(player.encorelives, 1)
			player.lives = #player.encorelives
		else
			player.lives = 0
		end
		E_Update(player)
		player.deadbypit = false
	end

	if player.lives > 0 then
		player.lives = #player.encorelives -- ok so there was a +1 here originally, however that would essentially mean you can cheat so no.
	end
end
addHook("PlayerThink", E_Main)

-- death
addHook("MobjDeath", function(target, inflictor, source, damagetype)
	-- detects if you are currently in Encore Mode
	if gametype ~= GT_ENCORE and multiplayer then
		return
	end
	-- whoops forgot there is no player, welp here is the player
	local player = target.player
	-- you are in Encore Mode right now, yay
	-- check if you are ok
	if not (player and player.valid and player.mo and player.mo.valid) then
		return
	end
	-- basically all this Hook does is check if you died because of a pit.
	if damagetype == DMG_DEATHPIT then
		player.deadbypit = true
	end
end, MT_PLAYER)

-- Prevents the player from respawning before Encore attempts to remove the character or game over.
addHook("PlayerCmd", function(player, cmd)
	if gametype ~= GT_ENCORE and multiplayer then
		-- WIP (TODO: remove bots, reset everything)
		return
	end
	
	if player.deadtimer and player.deadtimer <= 2*TICRATE then
		cmd.buttons = $ & ~BT_JUMP
	end
end)

-- HUD
hud.add(function(v, player)
	-- detects if you are currently in Encore Mode
	if gametype ~= GT_ENCORE and multiplayer then
		return
	end
	
	if (maptol & TOL_NIGHTS) then
		return
	end

	if not (player and player.valid and player.mo and player.mo.valid) then
		return
	end
	-- remove the regular lives HUD (CURSE YOU MODERN SONIC)
	hud.disable("lives")
	-- loopa
	for i in ipairs(player.encorelives) do
		if i <= 2 then
			v.drawScaled((16 + (i-1)*20)*FRACUNIT, 176*FRACUNIT, FRACUNIT/2, v.getSprite2Patch(player.encorelives[i].skin, SPR2_XTRA, A), V_SNAPTOBOTTOM|V_SNAPTOLEFT, v.getColormap(player.encorelives[i].skin, player.encorelives[i].color))
		elseif i <= 8 -- extended HUD
			v.drawScaled((56 + 9 + (i-3)*16)*FRACUNIT, (176+9+4)*FRACUNIT, 3*FRACUNIT/4, v.getSprite2Patch(player.encorelives[i].skin, SPR2_LIFE, A), V_SNAPTOBOTTOM|V_SNAPTOLEFT, v.getColormap(player.encorelives[i].skin, player.encorelives[i].color))
		else -- too many characters
			v.drawString(56+(i-3)*16, 176+8, "+"..tostring(#player.encorelives-8), V_SNAPTOBOTTOM|V_SNAPTOLEFT)
			break
		end
	end
end, "game")
